import shlex
import subprocess
import sys
from pathlib import Path
from typing import Any, override

import attrs
import git
import msgspec

from liblaf.cherries import core
from liblaf.cherries.typed import PathLike


def enc_hook(obj: Any) -> Any:
    if isinstance(obj, Path):
        return str(obj)
    raise NotImplementedError


@attrs.define
class Git(core.Run):
    inputs: list[Path] = attrs.field(factory=list)
    outputs: list[Path] = attrs.field(factory=list)
    repo: git.Repo = attrs.field(default=None)

    @override
    @core.impl(after=("Dvc",))
    def end(self, *args, **kwargs) -> None:
        if not self.repo.is_dirty(untracked_files=True):
            return
        self.repo.git.add(all=True)
        subprocess.run(["git", "status"], check=False)
        message: str = self._make_commit_message()
        self.repo.git.commit(message=message)

    @override
    @core.impl
    def log_input(self, path: PathLike, *args, **kwargs) -> None:
        path: Path = Path(path)
        self.inputs.append(path.relative_to(self.repo.working_dir))

    @override
    @core.impl
    def log_output(
        self,
        path: PathLike,
        name: PathLike | None = None,
        **kwargs,
    ) -> None:
        path: Path = Path(path)
        self.outputs.append(path.relative_to(self.repo.working_dir))

    @override
    @core.impl
    def start(self, *args, **kwargs) -> None:
        self.repo = git.Repo(self.plugin_root.exp_dir, search_parent_directories=True)

    def _make_commit_message(self) -> str:
        name: str = self.plugin_root.name
        message: str = f"chore(cherries): {name}\n\n"
        metadata: dict[str, Any] = {}
        metadata["cmd"] = shlex.join(sys.orig_argv)
        if url := self.plugin_root.url:
            metadata["url"] = url
        if params := self.plugin_root.get_params():
            metadata["params"] = params
        if inputs := self.inputs:
            metadata["inputs"] = inputs
        if outputs := self.outputs:
            metadata["outputs"] = outputs
        message += msgspec.yaml.encode(metadata, enc_hook=enc_hook).decode()
        return message



from collections.abc import Mapping
from pathlib import Path
from typing import Any, override

import attrs
import comet_ml
import cytoolz as toolz

from liblaf import grapes
from liblaf.cherries import core, paths
from liblaf.cherries.typed import PathLike


@attrs.define
class Comet(core.Run):
    disabled: bool = attrs.field(default=False)
    exp: comet_ml.CometExperiment = attrs.field(default=None)

    @override
    @core.impl(after=("Logging",))
    def end(self, *args, **kwargs) -> None:
        return self.exp.end()

    @override
    @core.impl
    def get_others(self) -> Mapping[str, Any]:
        return self.exp.others

    @override
    @core.impl
    def get_params(self) -> Mapping[str, Any]:
        return self.exp.params

    @override
    @core.impl
    def get_url(self) -> str:
        return self.exp.url  # pyright: ignore[reportReturnType]

    @override
    @core.impl(after=("Dvc",))
    def log_asset(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        path = Path(path)
        name = paths.as_posix(name)
        dvc_file: Path = path.with_name(path.name + ".dvc")
        if dvc_file.exists():
            path = dvc_file
            dvc_meta: Mapping[str, Any] = grapes.yaml.load(dvc_file)
            metadata = toolz.merge(metadata or {}, dvc_meta["outs"][0])
        self.exp.log_asset(path, name, metadata=metadata, **kwargs)  # pyright: ignore[reportArgumentType]

    @override
    @core.impl(after=("Dvc",))
    def log_input(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        if name is None:
            path = Path(path)
            name = f"inputs/{path.name}"
        metadata = toolz.assoc(metadata or {}, "type", "input")
        self.log_asset(path, name, metadata=metadata, **kwargs)

    @override
    @core.impl
    def log_metric(self, *args, **kwargs) -> None:
        return self.exp.log_metric(*args, **kwargs)

    @override
    @core.impl
    def log_metrics(self, *args, **kwargs) -> None:
        return self.exp.log_metrics(*args, **kwargs)

    @override
    @core.impl
    def log_other(self, *args, **kwargs) -> None:
        return self.exp.log_other(*args, **kwargs)

    @override
    @core.impl
    def log_others(self, *args, **kwargs) -> None:
        return self.exp.log_others(*args, **kwargs)

    @override
    @core.impl(after=("Dvc",))
    def log_output(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        if name is None:
            path = Path(path)
            name = f"outputs/{path.name}"
        metadata = toolz.assoc(metadata or {}, "type", "output")
        self.log_asset(path, name, metadata=metadata, **kwargs)

    @override
    @core.impl
    def log_parameter(self, *args, **kwargs) -> None:
        return self.exp.log_parameter(*args, **kwargs)

    @override
    @core.impl
    def log_parameters(self, *args, **kwargs) -> None:
        return self.exp.log_parameters(*args, **kwargs)

    @override
    @core.impl(after=("Logging",))
    def start(self, *args, **kwargs) -> None:
        self.exp = comet_ml.start(
            project_name=self.plugin_root.project_name,
            experiment_config=comet_ml.ExperimentConfig(
                disabled=self.disabled, name=self.plugin_root.name
            ),
        )



import contextlib
import datetime
import functools
from collections.abc import Mapping
from pathlib import Path
from typing import Any

import attrs
import git

from liblaf.cherries import paths
from liblaf.cherries.typed import PathLike

from ._plugin import Plugin
from ._spec import spec


@attrs.define
class Run(Plugin):
    """.

    References:
        1. [Experiment - Comet Docs](https://www.comet.com/docs/v2/api-and-sdk/python-sdk/reference/Experiment/)
        2. [Logger | ClearML](https://clear.ml/docs/latest/docs/references/sdk/logger)
        3. [MLflow Tracking APIs | MLflow](https://www.mlflow.org/docs/latest/ml/tracking/tracking-api/)
    """

    @functools.cached_property
    def data_dir(self) -> Path:
        if self._plugin_parent is not None:
            return self.plugin_root.data_dir
        return paths.data()

    @functools.cached_property
    def entrypoint(self) -> Path:
        if self._plugin_parent is not None:
            return self.plugin_root.entrypoint
        return paths.entrypoint()

    @functools.cached_property
    def exp_dir(self) -> Path:
        if self._plugin_parent is not None:
            return self.plugin_root.exp_dir
        return paths.exp_dir()

    @functools.cached_property
    def name(self) -> str:
        if self._plugin_parent is not None:
            return self.plugin_root.name
        return self.start_time.strftime("%Y-%m-%dT%H%M%S")

    @property
    def params(self) -> Mapping[str, Any]:
        if self._plugin_parent is not None:
            return self.plugin_root.params
        return self.get_params()

    @functools.cached_property
    def project_name(self) -> str | None:
        if self._plugin_parent is not None:
            return self.plugin_root.project_name
        try:
            repo: git.Repo = git.Repo(search_parent_directories=True)
        except git.InvalidGitRepositoryError:
            return None
        else:
            return Path(repo.working_dir).name

    @functools.cached_property
    def root_dir(self) -> Path:
        if self._plugin_parent is not None:
            return self.plugin_root.root_dir
        try:
            repo: git.Repo = git.Repo(search_parent_directories=True)
        except git.InvalidGitRepositoryError:
            return self.entrypoint.parent
        else:
            return Path(repo.working_dir).absolute()

    @functools.cached_property
    def start_time(self) -> datetime.datetime:
        if self._plugin_parent is not None:
            return self.plugin_root.start_time
        return datetime.datetime.now().astimezone()

    @functools.cached_property
    def url(self) -> str:
        if self._plugin_parent is not None:
            return self.plugin_root.url
        return self.get_url()

    @spec
    def end(self, *args, **kwargs) -> None: ...

    @spec(first_result=True)
    def get_others(self) -> Mapping[str, Any]: ...

    @spec(first_result=True)
    def get_params(self) -> Mapping[str, Any]: ...

    @spec(first_result=True)
    def get_url(self) -> str: ...

    @spec(delegate=False)
    def log_asset(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        if name is None:
            path = Path(path)
            with contextlib.suppress(ValueError):
                name = path.relative_to(self.data_dir)
        self.delegate("log_asset", (path, name), {"metadata": metadata, **kwargs})

    @spec
    def log_input(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None: ...

    @spec
    def log_metric(
        self,
        name: str,
        value: Any,
        /,
        step: int | None = None,
        epoch: int | None = None,
        **kwargs,
    ) -> None: ...

    @spec
    def log_metrics(
        self,
        dic: Mapping[str, Any],
        /,
        prefix: str | None = None,
        step: int | None = None,
        epoch: int | None = None,
        **kwargs,
    ) -> None: ...

    @spec
    def log_other(self, key: Any, value: Any, /, **kwargs) -> None: ...

    @spec
    def log_others(self, dictionary: Mapping[Any, Any], /, **kwargs) -> None: ...

    @spec
    def log_output(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None: ...

    @spec
    def log_parameter(
        self, name: Any, value: Any, /, step: int | None = None, **kwargs
    ) -> None: ...

    @spec
    def log_parameters(
        self,
        parameters: Mapping[Any, Any],
        /,
        prefix: str | None = None,
        step: int | None = None,
        **kwargs,
    ) -> None: ...

    @spec(delegate=False)
    def start(self, *args, **kwargs) -> None:
        self._plugins_prepare()
        self.delegate("start", args, kwargs)


active_run: Run = Run()
end = active_run.end
log_asset = active_run.log_asset
log_input = active_run.log_input
log_metric = active_run.log_metric
log_metrics = active_run.log_metrics
log_other = active_run.log_other
log_others = active_run.log_others
log_output = active_run.log_output
log_parameter = active_run.log_parameter
log_parameters = active_run.log_parameters
start = active_run.start



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import wrapt

if TYPE_CHECKING:
    from ._plugin import Plugin


def delegate_property_to_root[C: Callable](func: C) -> C:
    @wrapt.decorator
    def wrapper(
        wrapped: Callable, instance: "Plugin", args: tuple, kwargs: dict[str, Any]
    ) -> None:
        # TODO: make it work with `@functools.cached_property`
        if instance.plugin_root is not instance:
            return wrapped(*args, **kwargs)
        return getattr(instance.plugin_root, wrapped.__name__)

    return func
